# K funs UNIVARIATE.R    February 18, 2022    Author: Dwight Kincaid, PhD
#                                          dwight.kincaid@lehman.cuny.edu
#
#  SOURCE this file; its new funs can then be called. Some funs call 
#  other funs defined in this file. Missing values in the data are OK. 
#
#  These funs only deal with UNIVARIATE data. Commented-out DEMO CODE after 
#  each fun definition should be examined and run. If printed it's 24 pages.
#
#  DEMO DATA includes the N=298, data(IgM) in library(ISwR) as well as 
#  simulated data by sampling from theoretical probability distributions. 
#
#  I use base R funs as much as possible, to reduce reliance on contributed 
#  CRAN packages for reasons, including code transparency for students and 
#  for code 'survival' albeit at slightly slower speed for computationally 
#  intensive tasks although this is trivial unless N is large.  
# ---------------------------------------------------------------------------
##
##  FUNCTION  DEFINITIONS  --  for use on a single, numeric sample: a vector
##                                                  missing values OK
#      NEW FUNCTIONS                                                          
# 
#   1. freq.table(y, roundoff, ...)        # frequency distribution table                                  
#
#   2. EDA(y, data.name)                   # 4 EDA graphs, heavily annotated                                                       
#
#   3. CV.percent(y)                       # coefficient of variation as %
#
#   4. classical.CI.mean(y)           # 90,95,99% CI of mean by normal theory
#   5. boot.CI.mean(y, NS, data.name) # nonparametric bootstrap CI of mean
#
#   6. my.stats(y)                                  # descriptive stats 
#   7. new.stats(y, roundoff, data.name, norm.test) # descriptive stats & more
#
#   8. normal.QQ.plot(y)                 # graphical assessment
#   9. Normal.kernel.band(y, NS=1000)    #    of normality based on 
#                                        #       simulation from N(n, mean, sd)
#
#  10. %mc.skew%  # a binary operator: Monte Carlo sim test for SKEW     Ho: skw=0
#  11. %mc.kurt%  # a binary operator; Monte Carlo sim test for KURTOSIS Ho: krt=0
#
#  12. %boot.skew%  # a binary operator: nonparametric bootstrap CI for SKEW
#  13. %boot.kurt%  # a binary operator: nonparametric bootstrap CI for KURTOSIS
#                       
#  14. Tukey.outliers(y) # identify outliers by Tukey's 1.5 IQR rule in boxplot()
#
#  15. %bootMedian95CI%  # a binary operator; percentile bootstrap CI of median
#
#  16. skw(y)        # skewness (type 3); Kincaid code in base R
#  17. skw2(y)       # same as skw() but faster 
#
#  18. krt(y)        # kurtosis (type 3); Kincaid code in base R
#  19. krt2(y)       # same as krt() but faster
#
#  --  DEMO functions --
#
#  20. DEMO.What.is.SEM()   21. DEMO.SD.innate.biol.variability 
#
#  --  purely convenience functions  --
#  xx. thick.line(N)      # draw line of "="                                            
#  xx. thin.line(N)       # draw line of "-"
#  xx. my.pause()         # pause between graphs
#  xx. mytick(nx=2, ny=2, tick.ratio=0.5)  # minor ticks to base R graphs
#
#  --  functions to add  --
#  xx. myBanner     xx. FooterHeader  xx. BIG DATA vectors
#  xx. leaveOneOut  xx. simUnivar     xx. runs tests
#  xx. GIF graphs   xx. animations    xx. more EDA    xx. Bayesian funs  



#                        ----- 1 -----

freq.table <- function(y=rnorm(1e2, 50, 10), roundoff=4, ... ){

  # -------------------------------------------------------------------
  #  ARGUMENTS     frequency distribution table for a numeric vector
  #
  #  y         numeric vector for frequency distribution table
  #            missing values OK
  #
  #  roundoff  default: 4 decimal places
  #
  #  ...       args to hist(), 
  #               e.g., breaks=22, breaks="sturges", "scott" or "fd"
  #
  #  Function RETURNS a data frame by EXTRACTION from hist() object  
  #  
  #  Note: for the bins, it's   > ... <=   left open, right closed
  #        function may be called devoid of arguments, as a DEMO
  #   
  #  AUTHOR: Dwight Kincaid, PhD   dwight.kincaid@lehman.cuny.edu
  # -------------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove NA, missing values

  a <- hist(y, plot=FALSE, ...) # to extract from hist() object
  width    <- a$mids[2] - a$mids[1]                # width of hist bins
  L.level  <- round(a$mids - (width/2), roundoff)  # vector: lower bounds
  U.level  <- round(a$mids + (width/2), roundoff)  # vector: upper bounds
  my.level <- paste(L.level, ",", U.level, sep="") # character vector

  N <- length(y) # N of response variable

  # frequency distribution table by extraction, from hist() object 'a'
  dfr <- data.frame(interval  = 1:length(a$mids),
                    class     = my.level,  
                    midpt     = a$mids, 
                    FREQ      = a$counts,
                    rel.freq  = round(a$counts / N, roundoff),
                    cum.f     = cumsum(a$counts), 
                    cum.rel.f = round(cumsum(a$counts / N), roundoff),
                    density   = round(a$density, roundoff))

  return(dfr)   # data frame object returned 

} # end function definition


# demo the function  ----------------------------------------

# freq.table()   # as a DEMO, call the function with no args

# my.freq <- freq.table()
# my.freq
# str(my.freq)

# freq.table(roundoff=6, breaks=8) 

# library(ISwR); data(IgM)   # N=298
# freq.table(IgM)
# freq.table( IgM, breaks="scott") # breaks arg passed to hist()
# 

# freq.table(IgM, breaks=c(0,1,2,3,4,5)) # breaks arg passed to hist()
#  interval class midpt FREQ rel.freq cum.f cum.rel.f density  
#1        1   0,1   0.5  237   0.7953   237    0.7953  0.7953
#2        2   1,2   1.5   55   0.1846   292    0.9799  0.1846
#3        3   2,3   2.5    5   0.0168   297    0.9966  0.0168
#4        4   3,4   3.5    0   0.0000   297    0.9966  0.0000
#5        5   4,5   4.5    1   0.0034   298    1.0000  0.0034

# freq.table(y=rnorm(1e6), roundoff=5)  # N= one million
# freq.table(y=rnorm(5e7), roundoff=8)  # BIG DATA --- N = 50 million!



#                        ----- 2 -----


EDA <- function (y=rnorm(sample(10:80, size=1)), data.name){

  # --------------------------------------------------------------------------  
  #  ARGUMENTS
  #
  #  y            numeric vector, one sample of data, NAs OK
  #  data.name    descriptive string, e.g., "IgM, g/L"; truncated if long
  #
  #  EDA() RETURNS a graph window of 4 heavily annotated graphs and when
  #  called with no args, is a DEMO. Requires library(DescTools) for Jarque 
  #  BeraTest() and LillieTest(). Fun inspired by QuantPsych::eda.uni()
  #
  #  EDA: Exploratory Data Analysis refers to statistical & graphical methods
  #  developed by John Tukey et al., for data exploration & visualization. 
  #  Today, the analyst does EDA to see, to probe, to learn the unexpected,
  #  to communicate, and to understand the data, as well as to enhance 
  #  the generation of research hypotheses.
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # --------------------------------------------------------------------------

  if(missing(y)) {data.name <- "Simulated data"} # denote sampling from N(0,1)
  if(!missing(y) & missing(data.name)) {data.name <- deparse(substitute(y))}
 
  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove NA, missing values
  
  if(nchar(data.name) > 27) data.name <- paste(substr(data.name,1,26),"...",sep="")
  require(DescTools) # for JarqueBeraTest() & LillieTest() for normality

  par(mfrow = c(2, 2))  # 2 rows by 2 columns of graphs

    hist(y, main = "Histogram", cex.main=1.2, border="white",
      col="lightblue", xlab = data.name, cex.axis=.8,
      sub=paste("skew=", round(skw2(y),3),  "  kurtosis=", 
          round(krt2(y),3)), cex.sub=.9, col.sub="blue")
 
    if(length(y) <= 1e3) rug(y, col="gray")  # rug under histogram if N is low

    mtext(paste("N=", length(y), " Mean=", round(mean(y),4), 
                " SD=", round(sd(y),4)), cex=.7, col="blue")

    plot(density(y, n=1024), main = "Kernel Density Estimate", xlab=data.name, 
         cex.main=1.2, cex.axis=.8, col="red", lwd=3 )    
    
    boxplot(y, horizontal = TRUE,
        col="wheat", xlab=data.name, main= "Boxplot", ylab="", boxwex=.9,
        sub=paste("IQR=", round(IQR(y),3),
                  " min=", round(min(y),5), " max=", round(max(y),5)), 
        cex.sub=.9, col.sub="blue", cex.axis=.8 )

    mtext(paste("hinges: 25th=", round(quantile(y,.25),3),  
                ", 75th=", round(quantile(y,.75),3)), cex=.75, col="blue")

    outliers <- length( boxplot(y, plot=FALSE)$out) # extract from boxplot object

    mtext(paste(" N of outliers=", outliers, "  >1.5*IQR from hinge"), 
          side=3, line=-1, adj=0, cex=.7, col="blue")

    mtext(paste("median=", round(median(y),4) ), 
          side=1, line=-1, cex=.8, col="blue")

    a <- DescTools::JarqueBeraTest(y, robust=TRUE, method="chisq")$p.value 
    if(a == 0) a <- "<2.2e-16"  # if it's zero then make it a string to print
   
    # ------------------------------ Normal QQ plot to visualize normality
    if(is.character(a)){
        qqnorm(y, col="blue", sub=paste("Robust Jarque-Bera test p", a), 
        ylab="Observed Quantiles", cex.sub=.9, col.sub="blue", cex.axis=.8 )}
    else 
        {qqnorm(y, col="blue", sub=paste("Robust Jarque-Bera test p=", 
        signif(DescTools::JarqueBeraTest(y, robust=T, method="chisq")$p.value,5)), 
        ylab="Observed Quantiles", cex.sub=.9, col.sub="blue", cex.axis=.8 )}

    # ------------------------------

    if(length(y) <=  5000) mtext(paste("Shapiro-Wilk test p=", 
        signif(shapiro.test(y)$p.value,5)), cex=.8, col="blue")  # base R

    if(length(y) > 5000) mtext(paste("Lillifors Kolmogorov-Smirnov test p=", 
        signif(DescTools::LillieTest(y)$p.value,5)), cex=.7, col="blue") 

    qqline(y, col="blue")  # add straight line to Normal QQ plot 
    
    mtext(paste(" Visualization of observed distribution\n",
      " relative to normal distribution.", 
      "\n The more linear, the more normal."),
      side=3, line=-2.5, adj=0, cex=.6, col="lightblue4")

  par(mfrow = c(1, 1))  # reset to default: 1 graph per page

} # end function definition


# demo the function  ----------------------------------------

# EDA()  # call fun with no args, several times, to see what happens

# EDA(rnorm(1e2, 50, 10), data.name="Sim data from N(50,10)")

# library(ISwR); data(IgM)  # N=298
# EDA(IgM)
# EDA(IgM, data.name="IgM, g/L")  
# EDA(IgM, data.name="The quick brown fox jumped over the lazy dog")



#                        ----- 3 -----

CV.percent <- function(y){
 
  # ---------------------------------------------------------------- 
  #  ARGUMENTS      a useful and easy 'first' function for students
  #
  #  y         numeric vector, missing values OK
  #
  #  Function RETURNS Coefficient of Variation as a percent
  #                              100 * sd/mean
  #
  #  In package DescTools, try CoefVar(x, conf.level=.95) which 
  #    returns CV with 95% CI by noncentral t-distribution.
  #
  #  Suggestion: Add arg to CV.percent for bootstrap 95% CI of CV
  #              and/or CI of CV by DescTools::CoefVar()
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[ !is.na(y) ]                    # eliminate missing values, NA
  abs(100 * sd(y)/mean(y))              # abs() avoids a negative percent

} # end fun definition



# demo the function  ----------------------------------------

# library(ISwR); data(IgM)   # N=298
# CV.percent(IgM) # 58.5%

# --------------------------------------------------
# probe CV in a sample from the normal distribution
# change sample size, n, from 'low' to 'high' and see what happens
# n <- 50
# y <- rnorm(n, 100, 10)
# CV.percent(y) # 9.84%
#
# DescTools::CoefVar(y, conf.level=.95) # 8.24, 12.38%
#
# how to get bootstrap 95% CI of observed CV -------
# NS <- 1e5     # bootstrap 95% CI of CV !
# g <- replicate(NS, CV.percent(sample(y, replace=TRUE)))
# hist(g)
# quantile(g, c(.025, .975)) # 8.12, 11.24%  <-  95% CI of CV


# yy <- c(1, 2, 4, NA, 88)  # can it handle missing values? YES
# CV.percent(yy)

# yy <- rnorm(100, 50, 5) # CV should be close to 10%, obviously
# CV.percent(yy) 

# yy <- rnorm(1e3, 50, 5); CV.percent(yy) # CV should be close to 10%
# yy <- rnorm(1e4, 50, 5); CV.percent(yy) 
# yy <- rnorm(1e6, 50, 5); CV.percent(yy) 




#                        ----- 4 -----


classical.CI.mean <- function(y){

  # ----------------------------------------------------------   
  #  ARGS    y     numeric vector, missing values OK
  #
  #  RETURNS 90, 95, 99% CI of mean using t-distribution  
  #  AUTHOR: Dwight Kincaid   dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  N  <- length(y); Mean <- mean(y);    Min <- min(y) # get stats
  SD <- sd(y);     SEM  <- SD/sqrt(N); Max <- max(y)
  skew <- skw2(y); kurtosis <- krt2(y) # fast Kincaid funs 

  LB.90 <- Mean - qt(p=.95, df=N-1) * SEM    # 90% CI of mean
  UB.90 <- Mean + qt(p=.95, df=N-1) * SEM    #    by t-distribution
  CI.90 <- cbind(LB.90, UB.90) # pour into matrix

  LB.95 <- Mean - qt(p=.975, df=N-1) * SEM   # 95% CI of mean
  UB.95 <- Mean + qt(p=.975, df=N-1) * SEM
  CI.95 <- cbind(LB.95, UB.95) # pour into matrix

  LB.99 <- Mean - qt(p=.995, df=N-1) * SEM   # 99% CI of mean
  UB.99 <- Mean + qt(p=.995, df=N-1) * SEM
  CI.99 <- cbind(LB.99, UB.99) # pour into matrix

  out <- rbind(CI.90, CI.95, CI.99)
  colnames(out) <- c("LB", "UB")
  rownames(out) <- c("90% CI", "95% CI", "99% CI")

  # return a list  
  list(Confidence.Interval.of.Mean.using.t.distribution=out, 
       Descriptive.Statistics=cbind(N, Mean, SD, SEM, 
                                    Min, Max, skew, kurtosis))
} # end fun definition


# demo the function  ----------------------------------------

# library(ISwR); data(IgM)   # N=298

# classical.CI.mean(IgM)
# $`Confidence.Interval.of.Mean.using.t.distribution`
#               LB        UB
# 90% CI 0.7581446 0.8478957
# 95% CI 0.7494963 0.8565440
# 99% CI 0.7325116 0.8735287
# 
#$Descriptive.Statistics
#       N      Mean        SD       SEM Min Max     skew kurtosis
#[1,] 298 0.8030201 0.4694982 0.0271973 0.1 4.5 2.541542 13.33989


# y <- rnorm(10, 50, 10)
# classical.CI.mean(y)

# y <- rnorm(1e6, 50, 10)
# classical.CI.mean(y)



#                        ----- 5 -----    revised 1/9/22


boot.CI.mean <- function(y, NS=1e3, data.name){   

  # --------------------------------------------------------------
  #  ARGS
  #
  #  y           numeric vector, missing values OK
  #  NS          N of bootstrap samples
  #  data.name   string, e.g., "IgM, g/L", as desired
  #
  #  A simple, percentile bootstrap procedure in base R, using
  #  replicate() as iterator, returns graph of bootstrap 
  #  distribution of the mean, annotated with results, plus text
  #  results to R Console. 
  #  AUTHOR: Dwight Kincaid, PhD   dwight.kincaid@lehman.cuny.edu
  # --------------------------------------------------------------

  if(missing(data.name)) {data.name <- deparse(substitute(y))}

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  # truncate 'data.name' if too long
  if(nchar(data.name) > 27){ 
     data.name <- paste(substr(data.name,1,24),"...",sep="")}

  Mean <- mean(y); N   <- length(y)  # get basic stats
  SD <- sd(y);     SEM <- SD/sqrt(N) 

  time <- proc.time()                                   # begin timer
    out <- replicate(NS, mean(sample(y, replace=TRUE))) # iterate
  time <- ((proc.time() - time)/60)[[3]]       # boot time in minutes

  # --- percentile confidence intervals        # bootstrap 95% CI of mean
  LB.95 <- quantile(out, .025)
  UB.95 <- quantile(out, .975)
  CI.95 <- c(LB.95, UB.95)

  # --- traditional 95% CI based on normal theory   # use t-distribution
  LB.95t <- Mean - qt(p=.975, df=N-1) * SEM
  UB.95t <- Mean + qt(p=.975, df=N-1) * SEM

  hist(out, col="lightblue", border="white",
    main="Bootstrap distribution of the means", 
    sub=paste("red line: observed mean =", signif(Mean,6)),
    xlab="Mean", col.sub="red", cex.sub=.8, font.lab=2)

  abline(v=Mean, col="red", lwd=2); abline(v=CI.95)

  mtext(data.name, side=3, adj=0, cex=.9)
  mtext("percentile bootstrap 95% CI of mean",side=1,line=2,adj=0,cex=.8)   
  mtext(paste("LB: ", signif(LB.95,6)), side=1, line=3, adj=0, cex=.8)
  mtext(paste("UB: ", signif(UB.95,6)), side=1, line=4, adj=0, cex=.8)
  
  mtext(paste("NS: ", NS, "resamples"), side=4, line=0, adj=0, cex=.8)
  mtext(paste("minutes: ", signif(time,6)), side=4, line=0, adj=1, cex=.8)
  
  mtext("95% CI by normal theory", side=1, line=2, adj=1, cex=.8)
  mtext(paste("LB: ", signif(LB.95t,6)), side=1, line=3, adj=1, cex=.8)
  mtext(paste("UB: ", signif(UB.95t,6)), side=1, line=4, adj=1, cex=.8)

  LB.90 <- quantile(out,.05);  UB.90 <- quantile(out,.95)
  CI.90 <- c(LB.90,UB.90)

  LB.99 <- quantile(out,.005); UB.99 <- quantile(out,.995)
  CI.99 <- c(LB.99,UB.99)

  return(list("percentile bootstrap 90% CI of mean" = CI.90,
              "percentile bootstrap 95% CI of mean" = CI.95,
              "percentile bootstrap 99% CI of mean" = CI.99))

} # end function definition


# demo the function  ----------------------------------------

# library(ISwR); data(IgM)   # N=298
# boot.CI.mean(IgM, NS=1e4, data.name="IgM, g/L")
# classical.CI.mean(IgM)

# boot.CI.mean(IgM, NS=1e5)

# boot.CI.mean(rnorm(1e2))
# boot.CI.mean(rnorm(1e3), NS=1e5)



#                        ----- 6 -----


my.stats <- function(y){

  # ---------------------------------------------------------------  
  #  ARGS     y    numeric vector, missing values OK
  #
  #  Function RETURNS desriptive stats in a column
  #  AUTHOR: Dwight Kincaid, PhD   dwight.kincaid@lehman.cuny.edu
  # ---------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  # create 10 x 1 matrix
  A <- rbind(length(y), mean(y), sd(y), 
             min(y), max(y), skw2(y), krt2(y), median(y),
             sd(y)/sqrt(length(y)), IQR(y) )

  colnames(A) <- "descriptive stats"

  rownames(A) <-c("N", "Mean", "SD", "Min", "Max", 
                  "skew", "kurtosis", "Median", "SEM", "IQR")
  return(A)  

} # end fun definition


# demo the function  ----------------------------------------

# my.stats(rnorm(10))         # 10 from N(0,1)
# my.stats(rnorm(1e3))        # 1000 from N(0,1)
# my.stats(rnorm(1e6, 50, 5)) # a million from N(50,5)

# library(ISwR); data(IgM)   # N=298
# my.stats(IgM)



#                        ----- 7 -----


new.stats <- function(y, roundoff=3, data.name, norm.test=FALSE ){

  # ----------------------------------------------------------------
  #  ARGS
  #
  #  y            numeric vector, missing values OK  
  #  roundoff     decimal places, default 3
  #  data.name    short, descriptive string   
  #  norm.test    logical, default FALSE
  #
  #  requires packages:  moments, nortest
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  if(missing(data.name)) {data.name <- deparse(substitute(y))}

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                     # remove missing values
  if(length(y) < 8) norm.test=FALSE     # don't test normality on small sample

  yy <- y + rnorm(length(y), 0, .00001) # white noise to break ties in ks.test()
  
  N <- length(y); Mean <- mean(y); SEM <- sd(y)/sqrt(N)
  LB.95 <- Mean - qt(p=.975, df=N-1) * SEM  # 95% CI of mean
  UB.95 <- Mean + qt(p=.975, df=N-1) * SEM

  A <- rbind(N, Mean, sd(y), 100*sd(y)/Mean, min(y), max(y), skw2(y), krt2(y), 
          SEM, LB.95, UB.95, median(y), quantile(y, .25), quantile(y, .75), IQR(y))
  A <- round(A, roundoff)
  if(data.name != "") colnames(A) <- data.name else colnames(A) <- ""
  rownames(A) <-c("N", "Mean", "SD", "CV%", "Min", "Max", "skew", "kurtosis", 
          "SEM", "LB.95%CI.mean", "UB.95%CI.mean", "Median", "25th", "75th", "IQR")

  # Kincaid's revision of cvm.test() in library(nortest) to kill warnings about p
  my.cvm.test <- function (x){    # Cramer-von Mises normality test  >?cvm.test
    x <- sort(x[complete.cases(x)])
    n <- length(x); if (n < 8) stop("sample size must be greater than 7")
    p <- pnorm((x-mean(x))/sd(x))
    W <- (1/(12*n) + sum((p-(2*seq(1:n)-1)/(2*n))^2))
    WW <- (1+0.5/n)*W
    if (WW < 0.0275) { pval <- 1 - exp(-13.953 + 775.5 * WW - 12542.61 * WW^2)}
       else if (WW < 0.051) {pval <- 1 - exp(-5.903 + 179.546*WW - 1515.29*WW^2)}
       else if (WW < 0.092) {pval <- exp(0.886 - 31.62 * WW + 10.897 * WW^2)}
       else if (WW < 1.1)   {pval <- exp(1.111 - 34.242 * WW + 12.832 * WW^2)}
       else { pval <- 7.37e-10} #warning("p-value < 7.37e-10 cannot be computed")
    return(pval)
  }  # end of my.cvm.test() function definition

  if(norm.test == TRUE){
    require(moments); require(nortest) # GOF tests
	
    if(N <= 5000 & N > 7){
      B <- rbind(shapiro.test(y)$p.value, jarque.test(y)$p.value, ad.test(y)$p.value,
             my.cvm.test(y), sf.test(y)$p.value, pearson.test(y)$p.value,
             agostino.test(y)$p.value, anscombe.test(y)$p.value,
             bonett.test(y)$p.value, ks.test(yy, "pnorm", mean(yy), sd(yy))$p.value,
             lillie.test(y)$p.value)

      colnames(B) <- "p-value"
      rownames(B) <- c("Shapiro-Wilk test","Robust Jarque-Berra test",
                    "Anderson-Darling test", "Cramer-von Mises test", 
                    "Shapiro-Francia test", "Pearson test", "D'Agostino test", 
                    "Anscombe-Glynn test","Bonett-Seier test", 
                    "Kolmogorov-Smirnov test", "Lillifors Kolmogorov-Smirnov test") 
        return(list(statistics=A, goodness.of.fit.tests.for.normality=B)) }


    if(N > 5000 & N <= 46340){
      B <- rbind(jarque.test(y)$p.value, ad.test(y)$p.value, cvm.test(y)$p.value,
           pearson.test(y)$p.value,agostino.test(y)$p.value,anscombe.test(y)$p.value,
           bonett.test(y)$p.value, ks.test(yy, "pnorm", mean(yy), sd(yy))$p.value,
           lillie.test(y)$p.value) 

      colnames(B) <- "p-value"
      rownames(B) <- c("Robust Jarque-Berra test", "Anderson-Darling test",
                      "Cramer-von Mises test", "Pearson test",
                      "D'Agostino test", "Anscombe-Glynn test","Bonett-Seier test", 
                      "Kolmogorov-Smirnov test", "Lillifors Kolmogorov-Smirnov test") 
      return(list(statistics=A, goodness.of.fit.tests.for.normality=B)) }

    if(N > 46340){
      B <- rbind(jarque.test(y)$p.value, ad.test(y)$p.value, cvm.test(y)$p.value,
             pearson.test(y)$p.value, anscombe.test(y)$p.value,
             bonett.test(y)$p.value, ks.test(yy, "pnorm", mean(yy), sd(yy))$p.value,
             lillie.test(y)$p.value)

      colnames(B) <- "p-value"
      rownames(B) <- c("Robust Jarque-Berra test", "Anderson-Darling test", 
                      "Cramer-von Mises test", 
                      "Pearson test", "Anscombe-Glynn test","Bonett-Seier test", 
                      "Kolmogorov-Smirnov test", "Lillifors Kolmogorov-Smirnov test") 
      return(list(statistics=A, goodness.of.fit.tests.for.normality=B)) }
   }
 
   else

   return(A)   # return just the stats

}  # end function definition


# demo the function  ----------------------------------------

# library(ISwR); data(IgM)
# new.stats(IgM)        # minimal call
# new.stats(IgM, data.name="IgM in children")
# new.stats(IgM, roundoff=4, data.name="IgM in children")
# new.stats(IgM, roundoff=2, norm.test=TRUE)
# new.stats(IgM, roundoff=2, data.name="IgM in children", norm.test=TRUE)
# new.stats(rnorm(50))

# new.stats(rnorm(35), roundoff=3, norm.test=TRUE) # run 20 times
# new.stats(rnorm(1e6), roundoff=2, norm.test=TRUE)



#                        ----- 8 -----


normal.QQ.plot <- function(y=rnorm(15)){

  # ----------------------------------------------------------------
  #  ARGS  graphical assessment of normality of a sample, prevails
  #
  #  y     numeric vector, missing values OK  
  #
  #  requires package: lattice  for  my.qreference()
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values
  fun.name <- "normal.QQ.plot"

  y <- sort(y); n <- length(y); m <- mean(y); s <- sd(y)  # get stats
  quantiles <- ((1:n) - .5) / n      # quantile of each obs value
  obs.z.score <- (y - m) / s         # z-scores for graphical convenience
  normal.z.score <- qnorm(quantiles) # normal distribution scores

  my.pause <- function(){
    if(interactive()) readline("Hit <Enter> to continue...");invisible()}

  thin.line <- function(N=25) cat(rep("-", N), sep="", "\n\n")  # print "-"

  my.qreference <- function (test=NULL, m=50, nrep=6, 
    distribution=function(x) qnorm(x, 
    mean=ifelse(is.null(test), 0, mean(test)), sd=ifelse(is.null(test), 
        1, sd(test))), seed=NULL, nrows=NULL, cex.strip=0.75, 
    xlab=NULL, ylab=NULL){

    if (!is.null(seed)) 
        set.seed(seed)
    if (!is.null(test)) {
        testnam <- deparse(substitute(test))
        m <- length(test)
        av <- mean(test)
        sdev <- sd(test)
        fac <- factor(c(rep(testnam, m), paste("reference", 
            rep(1:(nrep - 1), rep(m, (nrep - 1))))))
        fac <- relevel(fac, ref = testnam)
    }
    id <- NULL
    if (is.null(nrows)) 
        nrows <- floor(sqrt(nrep))
    ncols <- ceiling(nrep/nrows)
    if (is.null(test)) {
        xy <- data.frame(y = distribution(runif(m * nrep)), fac=factor(rep(1:nrep, 
            rep(m, nrep))), id = factor(rep(1, m * nrep)))
        colpch <- c("black")
        qq <- lattice::qqmath(~y | fac, data = xy, par.strip.text=list(cex=0), 
            distribution = distribution, layout = c(ncols, nrows), 
            xlab = "", ylab = "", aspect = 1, pch = 16)
    }
    else {
        if (length(test) > 0) {
            yy <- quantile(test, c(0.25, 0.75))
            xx <- distribution(c(0.25, 0.75))
            r <- diff(yy)/diff(xx)
        }
        xy <- data.frame(y = c(test, distribution(runif(m * (nrep - 
            1)))), fac = fac, id = factor(rep(1:2, c(m, m * (nrep - 
            1)))))
        colpch <- c("black")
        if (is.null(xlab)) 
            xlab <- paste("Quantiles of", deparse(substitute(distribution)))
        if (is.null(ylab)) 
            ylab <- ""
        qq <- lattice::qqmath(~y | fac, data = xy, layout = c(ncols, 
            nrows), groups = id, aspect = 1, xlab = xlab, ylab = "", 
            pch = 16, distribution = distribution, 
            par.strip.text = list(cex = cex.strip))
    }
    print(qq)
  }

  # K fun, base R only, inspired by sim.qqnorm() in packge: animation
  my.sim.qqnorm <- function(N, NS=10, delay=1){
    for (i in 1:NS) {
      x <- rnorm(N, m, s)
      qqnorm(x, ylab="Monte Carlo simulated sample from normal distribution",pch=16)
      qqline(x, col=2, lwd=2)
      #abline(0,1, lwd=2, col="red") # add identity line
      mtext(paste("sim", i, "out of", NS), side=1, line=4, adj=1, font=2, cex=1.5)
      mtext(paste(" N =", N, ",", "from normal distribution; same N as observed N"), 
        adj=0, line=0, font=2)
      mtext("deviation from linearity is in direction of non-normality ", 
        side=1, line=-1, adj=1, cex=.7)
      mtext(paste(" Shapiro-Wilk normality test on sim data, p =", 
            signif(shapiro.test(x)$p.value,5)), adj=0, line=-1, cex=.8)
      Sys.sleep(delay)
    }
  }

  qqnorm(y)
  qqline(y, col=2)

  #plot(normal.z.score, obs.z.score, 
  #  ylab="Observed z-scores", xlab="Normal z-scores", 
  #  cex.lab=1.2, font.lab=2, 
  #  main="Normal QQ plot with linear fit", cex.main=1.8
  #)

  #grid()
  #qqline(y, col=2, lwd=2)
  #abline(lm(obs.z.score ~ normal.z.score), lwd=2, col="red") # add OLS fit
  #abline(0, 1, lwd=2, col="red") # add identity line

  if(n <= 5000){ 
    p <- signif(shapiro.test(y)$p.value, 4)
    mtext(paste("Shapiro-Wilk test of normality, p =", p), cex=.9)}

  mtext(paste("skew =", round(skw2(y), 3)), cex=.8, line=-1 )
  mtext(paste("kurtosis =", round(krt2(y), 3)), cex=.8, line=-2 )
  mtext(paste("", " n =", n), cex=.8, line=-1, adj=0)
  mtext(paste("", " mean =", round(m, 4)), cex=.8, line=-2, adj=0)
  mtext(paste("", " SD =", round(s, 4)), cex=.8, line=-3, adj=0)

  mtext("z = (y - mean)/SD", side=1, line=4, adj=0, cex=.8)
  mtext("z-scores are the data", side=1, line=2, adj=1, cex=.8)
  mtext("converted into SD", side=1, line=3, adj=1, cex=.8)
  mtext("units above & below the mean", side=1, adj=1, line=4, cex=.8)
  mtext("deviation from linearity is in direction of non-normality ", side=1,
         line=-1, adj=1, cex=.7)

  cat("\n\n"); thin.line(70)
  cat("OUTPUT from the R function: ", fun.name, " by Dwight Kincaid, PhD\n\n")
  cat("The Call: "); print(match.call()); cat("\n")

  cat("The sample's stats:\n")
  cat("\tn =", n, ", mean =", m, ", SD =", s, "\n")
  cat("\tskew =", round(skw(y), 4), ", kurtosis =", round(krt(y), 4), "\n\n")
  if(n <= 5000) cat(paste("Shapiro-Wilk test of normality, p =", p, "\n\n"))
  cat("See the graphs.\n")

  my.pause() # -------------------
  my.sim.qqnorm(N=n, NS=10, delay=2)

  my.pause() # --------------------
  hist(y, col="gray", main="Histogram of observed data", xlab="Response", 
    cex.lab=1.2, font.lab=2)

  my.pause() # --------------------
  dens <- density(y, n=2^10 ) 
  plot(dens, main="Kernel density estimate of observed data", 
       lwd=.5, font.lab=2, cex.lab=1.2)
  polygon(dens, col="lightblue") # color-in the kernel density curve

  my.pause() # --------------------
  library(lattice)
  reps <- 10
  for(i in 1:reps){
    my.qreference(y, cex.strip=1, xlab="z-scores", ylab="" )
    title(main="Assessment of normality in a function\noriginally by John Maindonald",
      cex.main=1.1, col.main="red", 
    sub="blue: observed data,  y-axis: response,  x-axis: theoretical quantiles", 
      col.sub="blue", cex.sub=.8, font.sub=2)
    mtext(paste("sim set #", i, " out of ", reps, sep=""), side=1, line=2, 
          adj=1, font=2, cex=2, col="red" )
    my.pause()
  }

  #my.pause() # --------------------
  yy <- seq(min(y), max(y), length.out=500)
  norm.fit <- dnorm(yy, mean(y), sd(y))
  ylim <- range(c(0, norm.fit, hist(y, plot=F)$density ))

  hist(y, freq=F, ylim=ylim, xlab="Observed data", col="gray", cex.lab=1.2, 
       font.lab=2, main="Density histogram with normal fit")
  points( yy, norm.fit, type="l", lwd=2.5, col="blue")

  cat("Time stamp:", date(), "\n"); thin.line(70)

} # end function definition


# demo the function -----------------------------------


# normal.QQ.plot()  # minimal call

# normal.QQ.plot(IgM)  # library(ISwR); data(IgM)

# y <- rnorm(1e3)
# normal.QQ.plot(y)

# normal.QQ.plot(rweibull(50, 2.5, 1)) # sample from the Weibull distribution



#                        ----- 9 -----

Normal.kernel.band <- function(y, NS=1e3){

  # -------------------------------------------------------------------
  #  Univariate visual assessment of normality by kernel density
  #  band from many data sets simulated from normal distribution 
  #  of the same n, mean and sd as observed data, y. 
  #  Returns annotated graph.
  #
  #  ARGS     y   numeric vector, missing values OK  
  #          NS   number of sim data sets from normal distribution,
  #               each generating a kernel density curve; 1000 default
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # -------------------------------------------------------------------

  # arg & data checking
  stopifnot(is.numeric(y), length(y) > 1, is.numeric(NS)) 
  y <- y[!is.na(y)]     # remove missing values
  
  n  <- length(y); m <- mean(y); s <- sd(y) # get stats
  mn <- min(y);   mx <- max(y) 
  g  <- hist(y, plot=FALSE) # just for plotting info
 
  # estimate 'max.y' for graph
  out <- replicate(NS/4, max(density(rnorm(n, m, s))$y))
  max.y <- max(out, hist(y, plot=FALSE)$density, density(y)$y )

#  hist(y, freq=FALSE, ylim=range(c(0, density(y)$y, max.y)),   
#    col="slategray4", xlab="Response", main="", border="white", 
#    cex.axis=1.2, cex.lab=1.2, font.lab=2, las=1)

  hist(y, freq=FALSE, ylim=range(c(0, density(y)$y, max.y)),   
    col="white", xlab="Response", main="", border="white", 
    cex.axis=1.2, cex.lab=1.2, font.lab=2, las=1)

  mtext(paste("Kernel density estimate for each of NS =", NS,
    "random samples from normal distribution"), line=2, cex=.9) 
  mtext(paste("of same N, Mean and SD as observed sample ",
    "-- a VISUAL assessment of normality"), line=1, cex=.9)

  for(i in 1:NS){
   # a <- density(rnorm(n, m, s), from=mn, to=mx, n=512)
    a <- density(rnorm(n, m, s), 
         from=min(g$breaks), to=max(g$breaks), n=512)

    #lines(a, col="lightgray", lwd=1.5)}
    lines(a, col=gray(.7, alpha=.3), lwd=1)}

  #lines(density(y, from=min(y), to=max(y)), lwd=2, col=2) # add obs kernel
  lines(density(y, from=min(g$breaks), to=max(g$breaks)), lwd=2, col=2) # add obs kernel



  hist(y, freq=FALSE, add=T, axes=F, ylim=range(c(0, density(y)$y, max.y)),   
    col=gray(.7, alpha=.3), main="", border="black")

  mtext("GRAY: kernel density for sim data", col="gray",
        side=1, line=2.5, adj=0, cex=.8, font=2)
  mtext(paste("RED: kernel density for observed data, N=", n), col="red",
        side=1, line=4, adj=0, cex=.8, font=1)

  mtext("If observed kernel density is well within", 
    side=1, line=2, adj=1, cex=.85)
  mtext("the band of sim kernel density", side=1, adj=1, line=3, cex=.85)
  mtext("this is evidence of normality", side=1, adj=1, line=4, cex=.85)

  mtext(paste("Observed: ", "  ", " skew =", round(skw2(y), 2)), adj=1, 
              line=0, cex=.8, col="red")
  mtext(paste("kurtosis =", round(krt2(y), 2)), adj=1, 
    line=-.8, cex=.8, col="red")
  if(n < 5000){mtext(paste("Shapiro-Wilk, p =", 
                     signif(shapiro.test(y)$p.value, 3)), 
                     adj=1, line=-1.6, cex=.8, col="red")}
  mtext(paste("Kolmogorov-Smirnov test, p =", 
              signif(ks.test(y, "pnorm", mean(y), sd(y))$p.value, 3)), 
              adj=1, line=-2.5, cex=.8, col="red")

} # end function definition


# demo the function -----

# Normal.kernel.band(y=rnorm(20))
# Normal.kernel.band(y=rnorm(1e2), NS=600)

# library(ISwR); data(IgM)
# Normal.kernel.band(IgM)
# Normal.kernel.band(IgM, NS=2e3)
# shapiro.test(IgM)  # N=3 to N=5000 max
# Normal.kernel.band(rweibull(1e2, 2, 5), NS=1e4)

# Normal.kernel.band(rnorm(100), NS=1e3) # 'law of big numbers'
# Normal.kernel.band(rnorm(10))          # 'law of small numbers'




#                        ----- 10 -----

"%mc.skew%" <- function(y, NS){

  # ----------------------------------------------------------------
  #  ARGS    y  numeric vector    NS  number of samples
  #                                   simulated by Monte Carlo
  #
  #  'binary operator' for one-sided, Monte Carlo simulation 
  #  significance test for SKEWNESS, achieved by random sampling
  #  from the normal distribution of the same N as y
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  N <- length(y); M <- mean(y); SD <- sd(y); obs.skew <- skw2(y)

  time <- proc.time()
     out <- replicate(NS, skw2(rnorm(N)))
  time <- ((proc.time() - time)/60)[[3]]  # total time in minutes

  ifelse(obs.skew > 0, NGE <- length(out[out >= obs.skew]),
     NGE <- length(out[out <= obs.skew]))  # handles pos or neg skew

  p <- (NGE+1) / (NS+1)  # the standard formula, one-sided

  # graph null distribution of skew under MC simulation test -------
  hist(out, col="gray", xlab="Skewness, type 3", main=NULL,
    font.lab=2, cex.lab=1.2, xlim=range(out, obs.skew) )
  
  mtext("Distribution of SKEW in samples from normal distribution", 
    line=2, cex=1.2) 
  mtext("of same sample size as observed data", cex=1.2, line=1) 
  abline(v=obs.skew, col="red", lwd=2)
  mtext("red line: observed skew", side=3, line=0, col="red", cex=.9)
  mtext(paste("NS =", NS, "Monte Carlo simulations"), side=1, adj=1, 
    line=4, col="red", cex=.8)
  mtext(paste("observed skew =", round(obs.skew,4)), 
    side=1, adj=1, line=3, col="red", cex=.8)
  mtext("Monte Carlo simulation test of skewness", side=4, line=-0,cex=1)
  mtext(paste("p =", signif(p, 5), "(1)"), side=1, line=3, adj=0,cex=.8)    
  mtext(paste("NGE =", NGE), side=1, line=4, adj=0, cex=.8)    

  # returns a list
  list("observed.skew"=obs.skew, "NS"=NS, "NGE"=NGE, 
       "p.one.sided"=p, "minutes"=time)

} # end binary operator definition


# demo binary operator  --------------------------------

# library(ISwR); data(IgM)   # N=298
# IgM%mc.skew%1e4
# IgM%mc.skew%1.11e5  # sometimes an odd number of permutations helps p readability

# rnorm(50)%mc.skew%1e4




#                        ----- 11 -----


"%mc.kurt%" <- function(y, NS){

  # ----------------------------------------------------------------
  #  ARGS    y  numeric vector    NS  number of samples
  #                                   simulated by Monte Carlo
  #  
  #  'binary operator' for one-sided, Monte Carlo simulation 
  #  significance test for KURTOSIS, achieved by random sampling 
  #  from the normal distribution of the same N as y
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  N <- length(y); M <- mean(y); SD <- sd(y); obs.kurt <- krt2(y)

  time <- proc.time()
     out <- replicate(NS, krt2(rnorm(N)))
  time <- ((proc.time() - time)/60)[[3]]  # total time in minutes

  ifelse(obs.kurt > 0, NGE <- length(out[out >= obs.kurt]),
     NGE <- length(out[out <= obs.kurt])) # handles pos or neg kurtosis

  p <- (NGE+1) / (NS+1)  # the standard formula, one-sided

  # graph null distribution of kurtosis under MC simulation test ------
  hist(out, col="gray", xlab="Kurtosis, type 3", main=NULL,
    font.lab=2, cex.lab=1.2, xlim=range(out, obs.kurt) )
  
  mtext("Distribution of KURTOSIS in samples from normal distribution", 
    line=2, cex=1.2) 
  mtext("of same sample size as observed data", cex=1.2, line=1) 
  abline(v=obs.kurt, col="red", lwd=2)
  mtext("red line: observed kurtosis", side=3, line=0, col="red",cex=.9)
  mtext(paste("NS =", NS, "Monte Carlo simulations"), side=1, adj=1, 
    line=4, col="red", cex=.8)
  mtext(paste("observed kurtosis =", round(obs.kurt,4)), 
    side=1, adj=1, line=3, col="red", cex=.8)
  mtext("Monte Carlo simulation test of kurtosis", side=4, line=-0,cex=1)
  mtext(paste("p =", signif(p, 5), "(1)"), side=1, line=3, adj=0, cex=.8)    
  mtext(paste("NGE =", NGE), side=1, line=4, adj=0, cex=.8)    

  list("observed.kurtosis"=obs.kurt, "NS"=NS, "NGE"=NGE, 
     "p.one.sided"=p, "minutes"=time)

} # end binary operator definition


# demo binary operator  ----------------------

# library(ISwR); data(IgM)   # N=298
# IgM%mc.kurt%1e4
# IgM%mc.kurt%1.11e5

# rnorm(20)%mc.kurt%1e4 # sometimes an odd number of permutations helps p readability




#                        ----- 12 -----


"%boot.skew%" <- function(y, NS){

  # ---------------------------------------------------------------
  #  ARGS    y  numeric vector    NS  number of bootstrap samples
  #
  #  'binary operator' for 
  #  nonparametric, percentile bootstrap 95% CI of SKEWNESS
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ---------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values

  N <- length(y); M <- mean(y); SD <- sd(y); obs.skew <- skw2(y)

  time <- proc.time()
     out <- replicate(NS, skw2( sample(y, replace=TRUE) ))
  time <- ((proc.time() - time)/60)[[3]]  # total time in minutes

  # --- nonparametric bootstrap percentile CI, 95%
  LB.95 <- quantile(out, .025); UB.95 <- quantile(out, .975)

  # graph the bootstrap distribution of skew 
  hist(out, col="gray", xlab="Skewness", main=NULL,
    font.lab=2, cex.lab=1.2, xlim=range(out, obs.skew) )
  
  mtext("Bootstrap Distribution of SKEWNESS", line=2, font=2,cex=1.2) 
  mtext("Percentile Bootstrap Confidence Interval, nonparametric",
   line=1, cex=.9) 
  abline(v=obs.skew, col="red", lwd=2)
  mtext("red line: observed skew", side=3, line=0, col="red", cex=.9)
  mtext(paste("NS =", NS, "Bootstrap samples"), side=1, adj=1, 
    line=4, col="red", cex=.8)
  mtext(paste("observed skew =", round(obs.skew,4)), 
    side=1, adj=1, line=3, col="red", cex=.8)
  mtext("Percentile Bootstrap, 95% CI", 
    side=1, adj=0, line=2, cex=.8)
  mtext(paste("LB =", signif(LB.95, 5)), side=1, line=3, adj=0, cex=.8)    
  mtext(paste("UB =", signif(UB.95, 5)), side=1, line=4, adj=0, cex=.8)   

  list("observed.skew"=obs.skew, "NS"=NS, "LB.95"=LB.95, "UB.95"=UB.95, 
     "skew.of.boot.distribution"=skw2(out),
     "minutes"=time)

} # end binary operator definition


# demo binary operator  ---------------------------

# library(ISwR); data(IgM)   #N=298
# IgM%boot.skew%1e4
# IgM%boot.skew%1e5

# rnorm(20)%boot.skew%1e4
# rnorm(1e3)%boot.skew%1e4
# rnorm(1e4)%boot.skew%1e3



#                        ----- 13 -----


"%boot.kurt%" <- function(y, NS){

  # --------------------------------------------------------------
  #  ARGS    y  numeric vector    NS  number of bootstrap samples
  #
  #  'binary operator' for 
  #  nonparametric, percentile bootstrap 95% CI of KURTOSIS
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ---------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values 

  N <- length(y); M <- mean(y); SD <- sd(y); obs.kurt <- krt2(y)

  time <- proc.time()
     out <- replicate(NS, krt2( sample(y, replace=TRUE) ))
  time <- ((proc.time() - time)/60)[[3]]  # total time in minutes

  # --- nonparametric bootstrap percentile CI, 95%
  LB.95 <- quantile(out, .025); UB.95 <- quantile(out, .975)

  # graph the bootstrap distribution of kurtosis 
  hist(out, col="gray", xlab="Kurtosis", main=NULL,
    font.lab=2, cex.lab=1.2, xlim=range(out, obs.kurt) )
  
  mtext("Bootstrap Distribution of KURTOSIS", line=2, font=2, cex=1.2) 
  mtext("Percentile Bootstrap Confidence Interval, nonparametric", 
    line=1, cex=.9) 
  abline(v=obs.kurt, col="red", lwd=2)
  mtext("red line: observed kurtosis", side=3, line=0, col="red",cex=.9)
  mtext(paste("NS =", NS, "Bootstrap samples"), side=1, adj=1, 
    line=4, col="red", cex=.8)
  mtext(paste("observed kurtosis =", round(obs.kurt,4)), 
    side=1, adj=1, line=3, col="red", cex=.8)
  mtext("Percentile Bootstrap, 95% CI", 
    side=1, adj=0, line=2, cex=.8)
  mtext(paste("LB =", signif(LB.95, 5)), side=1, line=3, adj=0, cex=.8)    
  mtext(paste("UB =", signif(UB.95, 5)), side=1, line=4, adj=0, cex=.8)   

  list("observed.kurt"=obs.kurt, "NS"=NS, "LB.95"=LB.95, "UB.95"=UB.95, 
     "skew.of.boot.distribution"=skw2(out),
     "minutes"=time)

} # end binary operator definition


# demo binary operator  ---------------------------

# library(ISwR); data(IgM)   # N=298
# IgM%boot.kurt%1e4
# IgM%boot.kurt%1e5

# rnorm(20)%boot.kurt%1e4
# rnorm(1e3)%boot.kurt%1e3
# rnorm(1e4)%boot.kurt%1e3




#                        ----- 14 -----


Tukey.outliers <- function(y){

  # ----------------------------------------------------------------
  #  ARGS     Tukey's rule:  points beyond 1.5 IQR beyond a hinge 
  #                          are potential OUTLIERS
  #  y    numeric vector, missing values OK
  #
  #  Returns a list of the trimmed vector, 
  #  matrix of row numbers of outliers, and the values
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values 

  outlier.values <- boxplot(y, plot=FALSE)$out   # outlier values
  outlier.rows   <- which(y %in% outlier.values) # rows of outliers
  trimmed <- y[-outlier.rows]                    # delete outliers
  list( trimmed.vector = trimmed,    
        outliers = cbind(outlier.rows, outlier.values))

}  # end fun definition


# demo the function  ----------------------------------------

# library(ISwR); data(IgM)   # N=298
# Tukey.outliers(IgM)
#
# out <- Tukey.outliers(IgM)
# str(out)
# new.IgM <- out$trimmed.vector  # create vector without Tukey outliers
# hist(new.IgM)

# new.IgM%mc.skew%1e4    # compare to original data,    IgM%mc.skew%1e4
# new.IgM%mc.kurt%1e4    # compare to original data,    IgM%mc.kurt%1e4
# shapiro.test(new.IgM); DAAG::qreference(new.IgM)
# EDA(new.IgM)
# normal.QQ.plot(new.IgM)
# Normal.kernel.band(new.IgM, NS=2e3)



#                        ----- 15 -----


"%bootMedian95CI%" <- function(y, NS){   

  # ----------------------------------------------------------------
  #  ARGS    y  numeric vector    NS  number of bootstrap samples
  #
  #  'binary operator' for 
  #  a nonparametric percentile bootstrap 95% CI of MEDIAN
  #
  #  AUTHOR: Dwight Kincaid, PhD    dwight.kincaid@lehman.cuny.edu
  # ----------------------------------------------------------------

  stopifnot(is.numeric(y), length(y) > 1)  # argument checking
  y <- y[!is.na(y)]                        # remove missing values
  quantile(replicate(NS, median(sample(y, replace=TRUE))), c(.025, .5, .975))

} # end fun definition


# demo the function  ----------------------------------------

# library(ISwR); data(IgM)
# IgM%bootMedian95CI%1e4  # awkward as IgM has many repeating values!

# y <- rnorm(1e2, 50, 5); y%bootMedian95CI%1e4
# median(y)




#                        ----- 16 -----

# SKEWNESS, type 3; fast Kincaid code -----------------------
skw <- function(y){            
  y <- y - sum(y)/length(y)
  (sum(y^3) / length(y)) / sd(y)^3}


#                        ----- 17 -----

skw2 <- function(y){            # faster than skw()
  n <- length(y)                # get stats
  m <- sum(y)/n                 # mean
  d  <- y-m                     # deviates
  s  <- sqrt(sum(d*d)/(n-1))    # SD
  sum(d*d*d)/n /(s*s*s)}        # M3/s^3 is type 3 skewness


#                        ----- 18 -----

# KURTOSIS, type 3; fast Kincaid code -----------------------
krt <- function(y){              
  y <- y - sum(y)/length(y)
  (sum(y^4) / length(y)) / sd(y)^4 - 3}


#                        ----- 19 -----

krt2 <- function(y){             # faster than krt()
  n <- length(y)                 # get stats
  m <- sum(y)/n                  # mean
  d <- y-m                       # deviates
  s  <- sqrt(sum(d*d)/(n-1))     # SD
  sum(d*d*d*d)/n /(s*s*s*s) - 3} # type 3 kurtosis

# skw(IgM)
# krt(IgM)

# set.seed(123); y <- rnorm(2e7)
# skw(y)   # 0.0001991044
# skw2(y)  #    "
# krt(y)   # -0.002147576
# krt2(y)  #    "
# Skew(y)  # 0.0001991044    library(DescTools) C++
# Kurt(y)  # -0.002147576      "        "
# also agrees with skew(), kurtosi() from library(psych)

# gc()      # time the funs for skew, kurtosis
# timer <- proc.time()
#  skw2(y)
# timer <- proc.time() - timer
# timer

# skw()  .95, .95, .95    2e7                 
# skw2() .18, .19, .22    2e7   5.28x faster   

# krt()  .94, .95, .94   
# krt2   .22  .20  .22    2e7   4.28x faster

# psych::describe(y)  # checks with K funs on skew, kurtosis



#                        ----- 20 -----


DEMO.What.is.SEM <- function(sim="norm", n=350, NS=1e4){

  # ARGS  -----------------------------------------------------------
  # sim     string; default "norm", "weibull" or "unif"
  # n       sample size of 'observed' sample; default n=350; try n=25
  # NS      number of sim samples to acquire; default NS=1e4
  # Note:   As 'n' increases observed SEM gets closer to the
  #         empirical, SD of the means 
  # Author: Dwight Kincaid, PhD   dwight.kincaid@lehman.cuny.edu
  # -----------------------------------------------------------------

  stopifnot(sim == "norm" | sim == "weibull" | sim == "uniform" | sim == "unif")
  stopifnot(is.numeric(n), is.numeric(NS), is.character(sim))

  my.pause <- function(){
    if(interactive())readline("\nHit <Enter> to continue...");invisible()}

  skw2 <- function(y){          # Kincaid's extremely fast code
    n <- length(y)              # get stats
    m <- sum(y)/n               # mean
    d  <- y-m                   # deviates
    s  <- sqrt(sum(d*d)/(n-1))  # SD
    sum(d*d*d)/n /(s*s*s)}      # M3/s^3 is type 3 skewness

  krt2 <- function(y){             # Kincaid's extremely fast code
    n <- length(y)                 # get stats
    m <- sum(y)/n                  # mean
    d <- y-m                       # deviates
    s  <- sqrt(sum(d*d)/(n-1))     # SD
    sum(d*d*d*d)/n /(s*s*s*s) - 3} # type 3 kurtosis

  # simple, modal dialog box ------------------------
  a <- if(interactive())
     askYesNo("Do you want to Call the function:  DEMO.What.is.SEM", default=TRUE,
     prompts = getOption("askYesNo", gettext(c("Yes", "No"))))
  stopifnot(a == TRUE)  # fun continues if TRUE is clicked

  plot.new()  # blank graph window scaled 0,1 in x&y; for text to be written 

  mtext("running function:", line=2, adj=-.1, cex=1.2, col="black", font=2)
  mtext("DEMO.What.is.SEM( )", line=-2, adj=.3, cex=2.7, col="darkred", font=2)
  mtext("by  Dr. Kincaid", line=-4.5, adj=1, cex=1.2, col="black", font=2)

  mtext("The standard error of the mean, SEM=SD/sqrt(N), of an observed sample is",
        line=-7, adj=0, cex=1, col="black", font=1)
  mtext("supposed to be the SD of the DISTRIBUTION of the MEANS from repeated",
        line=-9, adj=0, cex=1, col="black", font=1) 
  mtext("acquisition of independent samples of equal N, from the same statistical population.",
        line=-11, adj=0, cex=1, col="black", font=1)
  mtext("BUT, IS THIS TRUE? -- Data simulation enables answers.",
        line=-13, adj=0, cex=1.3, col="blue", font=2) 

  a <- formalArgs(DEMO.What.is.SEM) # the names of the args
  mtext(paste("The Arguments: ", a[1], " ", a[2], " ", a[3]), line=-16, adj=0)

  g    <- formals(DEMO.What.is.SEM) # arg names with values
  str1 <- paste("sim=", "'",g[[1]],"'", ", ", 
                "n=", g[[2]], ",", " NS=", g[[3]], sep="")
  mtext(paste("Defaults : ", str1), line=-18, adj=0) 
  mtext(paste("The Call : ", " sim=", "'",sim, "'", ",", " n=", n, ", ", " NS=", NS, 
               sep=""), line=-20, adj=0)
  mtext("Function can be called without arguments", line=-22, adj=1, col="peru")

  mtext("Argument info --", line=-24, adj=0, cex=.8, font=2)
  mtext("sim: string; default 'norm' ; or enter 'weibull' or 'unif'", 
        line=-25, adj=0, cex=.8)
  mtext(paste("n: sample size of 'observed' data; default:", g[[2]]), 
        line=-26, adj=0, cex=.8)
  mtext(paste("NS: number of sim samples to acquire; default:", g[[3]]), 
        line=-27, adj=0, cex=.8)

  mtext(sessionInfo()$running, line=-28.5, adj=1, cex=1, col="peru")
  mtext(sessionInfo()$R.version$version.string, line=-29.5, adj=1, cex=1, col="peru")
  mtext(paste("time stamp: ", date()), side=1, adj=1.05, line=4, cex=.85, col="black")
  mtext(paste("Remember:  SEM measures the variability\nof the MEANS,",
        "not the variability of the DATA"),
        side=1, line=3.1, adj=-.1, cex=1.1, col="darkred")
  my.pause()

  m1 <- 40; s1 <- 5
  if(sim == "norm")    y <- rnorm(n, m1, s1) # sample from N(m1, s1)
  if(sim == "weibull") y <- rweibull(n, shape=1, scale=1) # sample from the Weibull
  if(sim == "unif")    y <- runif(n, min=0, max=1) # sample from uniform distribution

  m <- mean(y); s <- sd(y); sem <- sd(y)/sqrt(n) # observed stats

  hist(y, main=paste("Histogram of our 'observed' sample\nof n =", n))

  mtext(paste("Mean =", round(m, 3)),   line=0,  adj=1, cex=.8)
  mtext(paste("SD =",   round(s, 3)),   line=-.9, adj=1, cex=.8)
  mtext(paste("SEM =",  round(sem, 3)), line=-1.8, adj=1, cex=.8)

  mtext(paste("skew =",     round(skw2(y), 3)), side=1, line=3, adj=0)
  mtext(paste("kurtosis =", round(krt2(y), 3)), side=1, line=4, adj=0)
  
  if(n <= 5000) mtext(paste("Shapiro-Wilk test\np =", signif(shapiro.test(y)$p.value, 5)), 
                      side=1, line=3.5, adj=1)
  my.pause()

  # get many samples the same way 'observed' data 'y' was acquired
  # use replicate() as the iterator

  timer <- proc.time()
    if(sim == "norm")    out <- replicate(NS, {x <- rnorm(n, m1, s1) ; mean(x)})
    if(sim == "weibull") out <- replicate(NS, {x <- rweibull(n, shape=1, scale=1); mean(x)})
    if(sim == "unif")    out <- replicate(NS, {x <- runif(n, min=0, max=1); mean(x)})
  timer <- proc.time() - timer; timer <- timer[[3]] # seconds

  s.out <- sd(out)  # SD of the distribution of Means

  hist(out, main=NULL, col="lightblue", border="white", 
       xlab=paste(NS, "Means"), cex.lab=1.4, col.sub="red", sub=paste("observed n =", n))

  abline(v=m, col="red")
  mtext("red line: observed mean", side=1, line=3, adj=0, col="red")

  mtext(paste("SEM of obs. sample =", round(sem, 7)), 
        side=1, line=2, adj=1, col="red")
  mtext(paste("SD of means =", round(s.out,7)), 
        side=1, line=4, adj=1, col="black")
  mtext(paste(round(timer, 6), "seconds elapsed"), 
        side=1, line=4, adj=0, cex=.8, col="blue")

  mtext("The STANDARD ERROR of the MEAN is supposed", line=2.6, cex=1.2)
  mtext("to be the SD of the DISTRIBUTION of the MEANS", line=1.5, cex=1.2)
  mtext("IS THIS TRUE?", line=.2, cex=1.2, col="blue", font=2)

} # end fun definition


# demo the function  ----------------------------------------
# DEMO.What.is.SEM()
# DEMO.What.is.SEM(n=25) # what happens with smaller observed n?
# DEMO.What.is.SEM(NS=1e5)
# DEMO.What.is.SEM(sim="weibull")



#                        ----- 21 -----


DEMO.SD.innate.biol.variability <- function(){

  # ARGS  --------------------------------------------------------
  #
  # Call the function without any arguments. 
  # Call it multiple times and reflect on the results.
  #
  # Author: Dwight Kincaid, PhD   dwight.kincaid@lehman.cuny.edu
  # --------------------------------------------------------------

  # pause between graphs but with leading linefeed
  idle <- function(){
    if(interactive())readline("\nHit<Enter>to continue...");invisible()}

  # simple, modal dialog box ------------------------
  a <- if(interactive())
     askYesNo("Do you want to Call the function:  DEMO.SD.innate.biol.variability", 
              default=TRUE,
     prompts = getOption("askYesNo", gettext(c("Yes", "No"))))
  stopifnot(a == TRUE)  # fun continues if TRUE is clicked

  #set.seed(123)  # works well with 123 as random seed, if desired
  n1 <- 1e6  # 5e6
  n2 <- sample(c(20:70, rep(100, 8), rep(200, 8)), size=1)      # 30:75   5:40
  y.big   <- rnorm(n1, 100, 10)
  y.small <- rnorm(n2, 100, 10)

  sd.big     <- round(sd(y.big), 4)        # some descriptive stats
  sd.small   <- round(sd(y.small), 3)      # for our 2 samples
  mean.big   <- round(mean(y.big), 4)
  mean.small <- round(mean(y.small), 3)
  SEM.big    <- round(sd.big / sqrt(n1), 3)
  SEM.small  <- round(sd.small / sqrt(n2), 3)

  # kernel density estimates for each sample ---------------------
  dens.big   <- density(y.big,   n=512, from=min(y.big),   to=max(y.big))
  dens.small <- density(y.small, n=512, from=min(y.small), to=max(y.small))

  hist(y.big, freq=FALSE,         # freq=F specifies a density histogram
    col=NULL, border=NA, main=NULL, xlab="Response", 
    xlim=range(c(y.big, y.small)), font.lab=2, cex.lab=1.2,
    ylim=range(c(dens.big$y, dens.small$y))
  )

  mtext("Distribution of two samples from N(100, 10)", line=2,
    cex=1.5, font=2, col="blue")

  lines(dens.big, col="red", lty="dashed", lwd=3.5)
  lines(dens.small, col="black", lwd=3.5)

  mtext("LARGE sample", adj=0, line=0, col="red", cex=1.2, font=2  )
  mtext("SMALLER sample", side=3, adj=1, line=0, cex=1.2, font=2  )

  mtext(paste(" ", " dashed: n =", n1), adj=0, col="red", line=-1.5)
  mtext(paste(" ", "SD =", sd.big),     adj=0, col="red", line=-3.5, 
        font=2, cex=1.3)
  mtext(paste(" ", " Mean =",mean.big), adj=0, col="red", line=-6) 
  mtext(paste(" ", " SEM =", SEM.big),  adj=0, col="red", line=-7)

  mtext(paste("black: n =", n2 ),         adj=1, line=-1.5)
  mtext(paste(" ", "SD =", sd.small),     adj=1, line=-3.5, 
        font=2, cex=1.3)
  mtext(paste(" ", " Mean =",mean.small), adj=1, line=-6) 
  mtext(paste(" ", " SEM =", SEM.small),  adj=1, line=-7)

  mtext("Remember: \nparametric SD = 10", side=1, adj=0, 
        line=4, cex=1.35, font=2, col="blue")
  mtext("NOTE: the 'large' sample \n is essentially, the 'population'", 
        side=1, line=4, adj=1, cex=.8)

} # end fun definition

# demo the fun -----------------------
# Call the fun several times

# DEMO.SD.innate.biol.variability()



############################################ --------
#  definitions for convenience functions   # --------------
#  the names of which are self-explanatory # --------------
############################################ --------

thick.line <- function(N=50) cat(rep("=", N), sep="", "\n")  # prints "="

thin.line <- function(N=50) cat(rep("-", N), sep="", "\n")  # prints "-"

my.pause <- function(){if(interactive()) readline("Hit <Enter> to continue...")
                       invisible()}

mytick <- function (nx = 2, ny = 2, tick.ratio = 0.5){
  # a shameless capture/simplification of fun definition for minor.tick()  
  # from library(Hmisc) as some have issues with Hmisc on old machines

  ax <- function(w, n, tick.ratio) {
    range <- par("usr")[if (w == "x") 
        1:2
    else 3:4]
    tick.pos <- if (w == "x") 
        par("xaxp")
    else par("yaxp")
    distance.between.minor <- (tick.pos[2] - tick.pos[1])/tick.pos[3]/n
    possible.minors <- tick.pos[1] - (0:100) * distance.between.minor
    low.candidates <- possible.minors >= range[1]
    low.minor <- if (any(low.candidates)) 
        min(possible.minors[low.candidates])
    else tick.pos[1]
    possible.minors <- tick.pos[2] + (0:100) * distance.between.minor
    hi.candidates <- possible.minors <= range[2]
    hi.minor <- if (any(hi.candidates)) 
        max(possible.minors[hi.candidates])
    else tick.pos[2]
    axis(if (w == "x") 
        1
    else 2, seq(low.minor, hi.minor, by = distance.between.minor), 
        labels = FALSE, tcl = par("tcl") * tick.ratio)
  }
  if (nx > 1) ax("x", nx, tick.ratio = tick.ratio)
  if (ny > 1) ax("y", ny, tick.ratio = tick.ratio)
  invisible()
} # end fun definition


# demo the function  ----------------------------------------

# hist(rnorm(1e4))
# mytick(2, 2)

# mytick(4, 4)
# mytick(2, 2, tick.ratio=1)

# hist(IgM); mytick(2, 2)

# ------------------------------
# END of FUNCTION DEFINITIONS    for single numeric samples
# ------------------------------

# end.  K funs UNIVARIATE.R